// X-Ray - The Interactive Game of Hide-and-Seek
//
// (C) Copyright 1996 by Kerry C. Imming
//      
// xray.cpp
 
#include <owl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <bwcc.h>
#include <inputdia.h>
#include <mmsystem.h>

#include "xray.h"

// Imbed copywrite notice in executable.
static char *xray_notice = "X-Ray.  Copyright 1996 by Kerry C. Imming.";

// The waveform device opened.
WORD  wDeviceID = 0;
char  elementName[255];
BOOL  bRunning = FALSE;

MCI_GENERIC_PARMS  MciGenParm;
MCI_OPEN_PARMS	   MciOpenParm;
MCI_PLAY_PARMS	   MciPlayParm;
MCI_SET_PARMS	   MciSetParm;

char *INIFile = ".\\XRAY.INI";

const int START_SCORE = 4000;
const int SCORE_RAY = 100;
const int SCORE_SECOND = 5;
const int BOXSIZE = 32;
const int XLAST = 8;
const int YLAST = 8;
const int MAXPOP = 6;

POINT makepoint(int x, int y)
/* makepoint: make a point from x & y components.                     */
{
  POINT temp;

  temp.x = x;
  temp.y = y;
  return temp;
}


typedef unsigned char BYTE;
enum typelist {null=0, topedge, botedge, leftedge, rightedge, empty, full};
enum raytype {avail=0, hit, deflect, deflect2, reflection};

const int CURSOR_ORG   = 0;
const int CURSOR_LEFT  = 1;
const int CURSOR_RIGHT = 2; 
const int CURSOR_TOP   = 3;  
const int CURSOR_BOT   = 4;   
const int CURSOR_MARK  = 5;

struct BoxPos {
    POINT UL;
    POINT LR;
    int type;
    int Row, Col;
    int RayX, RayY, RayType;
    BOOL avail;
    int mark;
  };
struct BoxPos Boxes[XLAST+2][YLAST+2];

struct HighScore {
    int score;
    char name[25];
} HighScores[5]; 


_CLASSDEF(TApp)
class TApp : public TApplication
{
public:
	TApp( LPSTR name, HINSTANCE hInstance,
		  HINSTANCE hPrevInstance, LPSTR lpCmd,
		  int nCmdShow)
	        : TApplication(name, hInstance,
			       hPrevInstance, lpCmd, nCmdShow) {};
	virtual void InitMainWindow( void );
        virtual void InitInstance( void );
};



class TGameWindow;
typedef TGameWindow* PGameWindow;

_CLASSDEF(TGameWindow)
class TGameWindow : public TWindow
{
private:
	HBITMAP GameOver, ScoreBMP, RemainBMP;
	HBITMAP Reflect, Hit, Deflect, Deflect2, MarkA, MarkB, BoxBMP, BoxgBMP;
        HBITMAP BoxBall, BoxBallX, MarkBX;
	HCURSOR Cursor[6];
	int nObjects;
	int Level;
	int Current_cur;
	BOOL SoundOn, TimerOn, FoundINI;
	int Counter, Tenths, Score, nRemain, nMarks;
	BOOL IsGameOver, IsGameStarted, IsGameAborted;
public:
	TGameWindow( PTWindowsObject AParent, LPSTR Title );
	virtual void About( TMessage& Message ) = [ CM_FIRST + IDM_ABOUT ];
        virtual void Best_Scores ( TMessage& Message ) = [ CM_FIRST + IDM_BEST ];
	void GetParms();
	void PutParms();
	POINT FindBox(int x, int y);
	int IsEdge(int row, int col);
	void PASCAL DrawBox( HDC DC, int X, int Y, HBITMAP BitMap, int rot );
	void PASCAL DrawBMP( HDC DC, int X, int Y, HBITMAP BitMap);
	void DrawGameOver( HDC DC );
	virtual void GetWindowClass( WNDCLASS& WndClass );
	virtual void Options( TMessage& Message ) = [ CM_FIRST + IDM_OPTION ];
	int  raytrace(int *row, int *col);
	void InitBoxes();
	void DrawEdge(HDC DC, int x, int y);
	void DrawMark(HDC DC, int x, int y);
	void DrawBoard(HDC DC);
	virtual void Paint( HDC PaintDC, PAINTSTRUCT& PaintInfo );
	virtual void ResetGame( TMessage& Message ) =  [ CM_FIRST + IDM_GAMENEW ];
	virtual void SetupWindow( void );
	virtual void Stop( TMessage& Message )        = [ CM_FIRST + IDM_STOP ];
	virtual void IDMSound( TMessage& Message )    = [ CM_FIRST + IDM_SOUND ];
	virtual void IDMTimer( TMessage& Message )    = [ CM_FIRST + IDM_TIMER ];
	virtual void IDMHelp( TMessage& Message )     = [ CM_FIRST + IDM_HELP ];
	virtual void IDMHelpHelp( TMessage& Message ) = [ CM_FIRST + IDM_HELPHELP ];
	virtual void GameExit( TMessage& Message )    = [ CM_FIRST + IDM_EXIT ];
	void StopGame( void );
	virtual void WMDestroy( TMessage& Message )     = [WM_FIRST + WM_DESTROY];
 	virtual void WMMouseMove(TMessage& Message )    = [WM_FIRST + WM_MOUSEMOVE];
	virtual void WMLButtonDown( TMessage& Message ) = [WM_FIRST + WM_LBUTTONDOWN];
	virtual void WMTimer( TMessage& Message ) 	= [WM_FIRST + WM_TIMER];
	virtual void WMSize( TMessage& Message ) 	= [WM_FIRST + WM_SIZE];

        void StopMCI ();
        virtual void MciNotify (TMessage& Message) = [WM_FIRST + MM_MCINOTIFY];
	void TGameWindow::PlayWave ( int wave );

	void WriteScore( HDC DC );

	friend class TOptionDialog;
	friend class TBestScoresDlg;
};

_CLASSDEF(TOptionDialog)
class TOptionDialog : public TDialog
{
public:
	TOptionDialog( PTWindowsObject AParent, LPSTR AName ) :
		TDialog( AParent, AName ) {}
	virtual BOOL CanClose(void);
	virtual void SetupWindow( void );
	virtual void WMHScroll( TMessage& Message ) = [ WM_FIRST + WM_HSCROLL ];
};

// --------------- TOptionDialog ---------------

void TOptionDialog::SetupWindow( void )
{
  TWindowsObject::SetupWindow();
  SetScrollRange( GetDlgItem( HWindow, IDD_POPSB ), SB_CTL, 1,
                  MAXPOP, FALSE );
  SetScrollPos( GetDlgItem( HWindow, IDD_POPSB ), SB_CTL,
		PGameWindow( Parent )->nObjects, TRUE );
}

void TOptionDialog::WMHScroll( TMessage& Message )
{
  int Pos;
  HWND Scroll;
  const int PageStep = 1;

  Scroll = (HWND)Message.LP.Hi;
  Pos = GetScrollPos( Scroll, SB_CTL );
  switch ( Message.WParam )
  {
    case SB_LINEUP:
	Pos--;
	break;
    case SB_LINEDOWN:
	Pos++;
	break;
    case SB_PAGEUP:
	Pos -= PageStep;
	break;
    case SB_PAGEDOWN:
	Pos += PageStep;
	break;
    case SB_THUMBPOSITION:
    case SB_THUMBTRACK:
	Pos = Message.LP.Lo;
	break;
    }
  SetScrollPos( Scroll, SB_CTL, Pos, TRUE );
}

BOOL TOptionDialog::CanClose()
{

  PGameWindow( Parent )->nObjects =
   GetScrollPos( GetDlgItem( HWindow, IDD_POPSB ),SB_CTL);

  return TRUE;
}

// --------------- TBestScoresDlg ---------------

_CLASSDEF(TBestScoresDlg)
class TBestScoresDlg : public TDialog
{
public:
	TBestScoresDlg( PTWindowsObject AParent, LPSTR AName ) :
		TDialog( AParent, AName ) {}
	void Reset_Scores( TMessage& Message ) = [ ID_FIRST + IDM_RESET];
	void Update_Scores(void);
	virtual void SetupWindow( void );
};


void TBestScoresDlg::SetupWindow( void )
{
   TWindowsObject::SetupWindow();
   Update_Scores();
}

void TBestScoresDlg::Reset_Scores( TMessage& )
{
  for (int i=0;i<5;i++)
  {
    HighScores[i].score = 0;
    strcpy(HighScores[i].name, "Anonymous");
  }
  Update_Scores();
}

void TBestScoresDlg::Update_Scores( void )
{
   char str[25];

   sprintf( str, "%1d.  %5d", 1, HighScores[0].score);
   SetDlgItemText( HWindow, IDD_BEST1, str );
   sprintf( str, "%1d.  %5d", 2, HighScores[1].score);
   SetDlgItemText( HWindow, IDD_BEST2, str );
   sprintf( str, "%1d.  %5d", 3, HighScores[2].score);
   SetDlgItemText( HWindow, IDD_BEST3, str );    
   sprintf( str, "%1d.  %5d", 4, HighScores[3].score);
   SetDlgItemText( HWindow, IDD_BEST4, str );
   sprintf( str, "%1d.  %5d", 5, HighScores[4].score);
   SetDlgItemText( HWindow, IDD_BEST5, str );    

   sprintf( str, "%-24s", HighScores[0].name);
   SetDlgItemText( HWindow, IDD_NAME1, str );
   sprintf( str, "%-24s", HighScores[1].name);
   SetDlgItemText( HWindow, IDD_NAME2, str );
   sprintf( str, "%-24s", HighScores[2].name);
   SetDlgItemText( HWindow, IDD_NAME3, str );    
   sprintf( str, "%-24s", HighScores[3].name);
   SetDlgItemText( HWindow, IDD_NAME4, str );
   sprintf( str, "%-24s", HighScores[4].name);
   SetDlgItemText( HWindow, IDD_NAME5, str );    
}

// --------------- TGameWindow -----------------

TGameWindow::TGameWindow( PTWindowsObject AParent, LPSTR Title ) :
	TWindow( AParent, Title )
{
	Attr.W = 360;
	Attr.H = 400;
	Attr.Style = WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX;
	randomize();
	Current_cur = 0;
};

void TGameWindow::About( TMessage& )
{
  GetApplication()->ExecDialog(new TDialog(this, "About"));
}

void TGameWindow::Best_Scores( TMessage& )
{
  GetApplication()->ExecDialog(new TBestScoresDlg(this, "BESTSCORES"));
}

void TGameWindow::GetParms()
{
  int i;
  FILE *tstini;

  if ((tstini = fopen(INIFile, "rt")) == NULL)
    FoundINI = FALSE;
  else {
    FoundINI = TRUE;
    fclose(tstini);
  }

  Level  = GetPrivateProfileInt("XRAY", "Level", 1, INIFile);
  nObjects = GetPrivateProfileInt("XRAY", "Number", 4, INIFile);
  if (nObjects < 1) nObjects = 1;
  SoundOn = (BOOL)GetPrivateProfileInt("XRAY", "SoundOn", 0, INIFile);
  TimerOn = (BOOL)GetPrivateProfileInt("XRAY", "TimerOn", 1, INIFile);

  HighScores[0].score = GetPrivateProfileInt("XRAY", "Best1", 0, INIFile);
  HighScores[1].score = GetPrivateProfileInt("XRAY", "Best2", 0, INIFile);
  HighScores[2].score = GetPrivateProfileInt("XRAY", "Best3", 0, INIFile);
  HighScores[3].score = GetPrivateProfileInt("XRAY", "Best4", 0, INIFile);
  HighScores[4].score = GetPrivateProfileInt("XRAY", "Best5", 0, INIFile);

  GetPrivateProfileString("XRAY", "Name1","Anonymous",
    HighScores[0].name, sizeof(HighScores[0].name), INIFile);
  GetPrivateProfileString("XRAY", "Name2","Anonymous",
    HighScores[1].name, sizeof(HighScores[1].name), INIFile);
  GetPrivateProfileString("XRAY", "Name3","Anonymous",
    HighScores[2].name, sizeof(HighScores[2].name), INIFile);
  GetPrivateProfileString("XRAY", "Name4","Anonymous",
    HighScores[3].name, sizeof(HighScores[3].name), INIFile);
  GetPrivateProfileString("XRAY", "Name5","Anonymous",
    HighScores[4].name, sizeof(HighScores[4].name), INIFile);

  for (i=0; i<5; i++) {
    if (HighScores[i].score < 0 ||
	HighScores[i].score > START_SCORE)
      HighScores[i].score = 0;
  }

  if (!SoundOn)  // defaults to checked at startup
    CheckMenuItem(GetMenu( HWindow ), IDM_SOUND, MF_UNCHECKED);
  if (!TimerOn)  // defaults to checked at startup
    CheckMenuItem(GetMenu( HWindow ), IDM_TIMER, MF_UNCHECKED);
}

void TGameWindow::PutParms()
{
  char buf[80];

  if (!FoundINI) return;

  sprintf(buf, "%d", Level);
  WritePrivateProfileString("XRAY", "Level", buf, INIFile);

  sprintf(buf, "%d", nObjects);
  WritePrivateProfileString("XRAY", "Number", buf, INIFile);

  sprintf(buf, "%d", (int)SoundOn);
  WritePrivateProfileString("XRAY", "SoundOn", buf, INIFile);

  sprintf(buf, "%d", (int)TimerOn);
  WritePrivateProfileString("XRAY", "TimerOn", buf, INIFile);

  sprintf(buf, "%d", HighScores[0].score);
  WritePrivateProfileString("XRAY", "Best1", buf, INIFile);
  sprintf(buf, "%d", HighScores[1].score);
  WritePrivateProfileString("XRAY", "Best2", buf, INIFile);
  sprintf(buf, "%d", HighScores[2].score);
  WritePrivateProfileString("XRAY", "Best3", buf, INIFile);
  sprintf(buf, "%d", HighScores[3].score);
  WritePrivateProfileString("XRAY", "Best4", buf, INIFile);
  sprintf(buf, "%d", HighScores[4].score);
  WritePrivateProfileString("XRAY", "Best5", buf, INIFile);


  WritePrivateProfileString("XRAY", "Name1", HighScores[0].name, INIFile);
  WritePrivateProfileString("XRAY", "Name2", HighScores[1].name, INIFile);
  WritePrivateProfileString("XRAY", "Name3", HighScores[2].name, INIFile);
  WritePrivateProfileString("XRAY", "Name4", HighScores[3].name, INIFile);
  WritePrivateProfileString("XRAY", "Name5", HighScores[4].name, INIFile);
}

void PASCAL TGameWindow::DrawBMP( HDC DC, int X, int Y, HBITMAP BitMap)
{
  HDC MemDC;
  BITMAP bm;
  BOOL MadeDC;

  if ( DC == 0 )
  {
    DC = GetDC( HWindow );
    MadeDC = TRUE;
  }
  else
    MadeDC = FALSE;
  MemDC = CreateCompatibleDC( DC );
  SelectObject( MemDC, BitMap );
  GetObject( BitMap, sizeof( bm ), ( LPSTR ) &bm );

  BitBlt( DC, X, Y, bm.bmWidth, bm.bmHeight, MemDC, 0, 0, SRCCOPY );

  DeleteDC( MemDC );
  if ( MadeDC )
    ReleaseDC( HWindow, DC );
}

void PASCAL TGameWindow::DrawBox( HDC DC, int X, int Y, HBITMAP BitMap, int rot )
{
  HDC MemDC;
  BITMAP bm;
  BOOL MadeDC;

  if ( DC == 0 )
  {
    DC = GetDC( HWindow );
    MadeDC = TRUE;
  }
  else
    MadeDC = FALSE;
  MemDC = CreateCompatibleDC( DC );
  SelectObject( MemDC, BitMap );
  GetObject( BitMap, sizeof( bm ), ( LPSTR ) &bm );

  switch (rot)
  {
    case 90:
      BitBlt( DC, X, Y, BOXSIZE, BOXSIZE, MemDC, BOXSIZE, 0, SRCCOPY );
      break;
    case 180:
      BitBlt( DC, X, Y, BOXSIZE, BOXSIZE, MemDC, BOXSIZE*2, 0, SRCCOPY );
      break;
    case 270:
      BitBlt( DC, X, Y, BOXSIZE, BOXSIZE, MemDC, BOXSIZE*3, 0, SRCCOPY );
      break;
    case 0:
    default:
      BitBlt( DC, X, Y, BOXSIZE, BOXSIZE, MemDC, 0, 0, SRCCOPY );
  }

  DeleteDC( MemDC );
  if ( MadeDC )
    ReleaseDC( HWindow, DC );
}


void TGameWindow::DrawGameOver( HDC DC )
{
  int x,y;

  /* draw balls */
  for (y=0; y<=YLAST+1; y++)
    for (x=0; x<=XLAST+1; x++) {
      if (Boxes[x][y].type == full && Boxes[x][y].mark == 2)
	DrawBox( DC, Boxes[x][y].UL.x, Boxes[x][y].UL.y, BoxBall,0);
      else if (Boxes[x][y].type == empty && Boxes[x][y].mark == 2)
	DrawBox( DC, Boxes[x][y].UL.x, Boxes[x][y].UL.y, MarkBX,0);
      else if (Boxes[x][y].type == full && Boxes[x][y].mark != 2)
	DrawBox( DC, Boxes[x][y].UL.x, Boxes[x][y].UL.y, BoxBallX,0);
    }

  DrawBMP( DC, 105, 2, GameOver);
}


void TGameWindow::GetWindowClass( WNDCLASS& WndClass )
{
  TWindow::GetWindowClass( WndClass );
  WndClass.style = 0;
  WndClass.hbrBackground = (HBRUSH)GetStockObject( LTGRAY_BRUSH );
  WndClass.lpszMenuName = "XRAYMENU";
  WndClass.hIcon = LoadIcon( GetApplication()->hInstance, "XrayBD" );
}

void TGameWindow::Options( TMessage& Message )
{
  int i;

  i = nObjects;
  GetApplication()->ExecDialog(new TOptionDialog(this, "OptionDlg"));
  if (i != nObjects && !IsGameOver) {
    if (IsGameStarted) {
      if (MessageBox( HWindow, "Number of objects changed, restart game?",
	"X-Ray Message", MB_YESNO | MB_ICONQUESTION ) == IDYES) {
        IsGameAborted = TRUE;
	StopGame();
        ResetGame( Message );
      }
    }
    else {
      IsGameAborted = TRUE;
      StopGame();
      ResetGame( Message );
    }
  }
}

int TGameWindow::IsEdge(int row, int col)
/* Check if box position (row,col) is an edge  */
{
  return ( Boxes[row][col].type == topedge   ||
	   Boxes[row][col].type == botedge   ||
	   Boxes[row][col].type == rightedge ||
	   Boxes[row][col].type == leftedge );
}              

POINT TGameWindow::FindBox(int x, int y)
/* findbox: find the box in boxes array that contains point x,y */
/*          returns 0,0 if not found                            */
{
  int row,col;
  POINT temp;

  temp.x = 0;
  temp.y = 0;

  if (x < Boxes[0][0].UL.x) return temp;
  /* find the right col */
  for (col=0; col<=XLAST+1; col++)
  {  
    if (x < Boxes[col][0].LR.x)
    {
      /* In this col, now find the right row */
      if (y < Boxes[col][0].UL.y) return temp;
      for (row=0; row<=YLAST+1; row++)
      {
	if (y < Boxes[col][row].LR.y)
	{
	  temp.x = col;
	  temp.y = row;
	  return temp;  
	}
      }
    }
  }
  return temp;
}


void TGameWindow::DrawEdge(HDC DC, int x, int y)
{
  int rotate;

  if (Boxes[x][y].avail) {
    DrawBox(DC,Boxes[x][y].UL.x,Boxes[x][y].UL.y,BoxgBMP, 0);
    return;
  }

  if (Boxes[x][y].type == topedge)
    rotate = 90;
  else if (Boxes[x][y].type == rightedge)
    rotate = 180;
  else if (Boxes[x][y].type == botedge)
    rotate = 270;
  else
    rotate = 0;

  if (Boxes[x][y].RayType == hit)
    DrawBox(DC,Boxes[x][y].UL.x,Boxes[x][y].UL.y,Hit,rotate);
  else if (Boxes[x][y].RayType == deflect)
    DrawBox(DC,Boxes[x][y].UL.x,Boxes[x][y].UL.y,Deflect,rotate);
  else if (Boxes[x][y].RayType == deflect2)
    DrawBox(DC,Boxes[x][y].UL.x,Boxes[x][y].UL.y,Deflect2,rotate);
  else if (Boxes[x][y].RayType == reflection)
    DrawBox(DC,Boxes[x][y].UL.x,Boxes[x][y].UL.y,Reflect,rotate);

  return;
}

void TGameWindow::DrawMark(HDC DC, int x, int y)
{
  if (Boxes[x][y].mark > 0) {

    switch (Boxes[x][y].mark)
    {
      case 2:
	DrawBox(DC, Boxes[x][y].UL.x, Boxes[x][y].UL.y, MarkB, 0);
	break;
      case 3:
	DrawBox(DC, Boxes[x][y].UL.x, Boxes[x][y].UL.y, MarkA, 0);
	break;
      case 1:
		/* Erase mark */
      default:
	DrawBox(DC, Boxes[x][y].UL.x, Boxes[x][y].UL.y, BoxBMP, 0);
     }
   }
  return;
}

int TGameWindow::raytrace(int *row, int *col)
/* raytrace: trace from row,col to an edge */
{
  enum { horizontal=0, vertical };
  int i, dir, dirtype;
  int r, c;

  r = *row;
  c = *col;

  if      (Boxes[r][c].type == topedge)   { dir =  1; dirtype = vertical;   }
  else if (Boxes[r][c].type == botedge)   { dir = -1; dirtype = vertical;   }
  else if (Boxes[r][c].type == leftedge)  { dir =  1; dirtype = horizontal; }
  else if (Boxes[r][c].type == rightedge) { dir = -1; dirtype = horizontal; }
  else { r=-1; c=-1; return -1; }

  /* Special case, first move */
  if (dirtype == vertical) {
    if (Boxes[r][c+dir].type == full) {
       /* hit */
      return hit;
    } 
    else if (Boxes[r+1][c+dir].type == full) {
      return reflection;
    } 
    else if (Boxes[r-1][c+dir].type == full) {
      return reflection;
    }
    else c += dir;
  }
  if (dirtype == horizontal) {
    if (Boxes[r+dir][c].type == full) {
       /* hit */
      return hit;
    } 
    else if (Boxes[r+dir][c+1].type == full) {
      return reflection;
    } 
    else if (Boxes[r+dir][c-1].type == full) {
      return reflection;
    }   
    else r += dir; 
  }


  i = 1;
  while (i < 100 && !IsEdge(r,c)) {   
                       
    if (dirtype == vertical) {
      if (Boxes[r][c+dir].type == full) {
        return hit;  /* hit */
      }  
      else if (Boxes[r+1][c+dir].type == full) {
        dir = -1;
	dirtype = horizontal;
      }
      else if (Boxes[r-1][c+dir].type == full) {
        dir = 1;
	dirtype = horizontal;
      }
      else c += dir;
    }
    if (dirtype == horizontal) {
      if (Boxes[r+dir][c].type == full) {
         /* hit */
        return hit;
      } 
      else if (Boxes[r+dir][c+1].type == full) {
        dir = -1;
	dirtype = vertical;
      }
      else if (Boxes[r+dir][c-1].type == full) {
        dir = 1;
	dirtype = vertical;
      }
      else r += dir; 
    }

  i++;
  }

  if (*row == r && *col == c) return reflection;
  else {
    *row = r;
    *col = c;
    return deflect;
  }
}


void TGameWindow::DrawBoard(HDC DC)
{
  int x,y;

  /* draw boxes */
  for (y=0; y<=YLAST+1; y++)
    for (x=0; x<=XLAST+1; x++)
      if (IsEdge(x,y))
      	DrawEdge(DC,x,y);
      else
        DrawMark(DC,x,y);

  return;
}

void TGameWindow::InitBoxes()
{
  int i,x,y;

  /* Set Row, Col for All boxes */
  for (y=0; y<=YLAST+1; y++)
    for (x=0; x<=XLAST+1; x++)
    {
      Boxes[x][y].UL = makepoint( x   *BOXSIZE+20, y   *BOXSIZE+24);
      Boxes[x][y].LR = makepoint((x+1)*BOXSIZE+20,(y+1)*BOXSIZE+24);
      Boxes[x][y].type = empty;
      Boxes[x][y].Row = x;
      Boxes[x][y].Col = y;
      Boxes[x][y].RayX = 0;
      Boxes[x][y].RayY = 0;
      Boxes[x][y].RayType = avail;
      Boxes[x][y].avail = FALSE;
      Boxes[x][y].mark = 0;
    }

   /* Set Mark Value for inner boxes */
  for (y=1; y<=YLAST; y++)
    for (x=1; x<=XLAST; x++)
    {
      Boxes[x][y].mark = 1;
    }

  /* initialize row 0 */
  Boxes[0][0].type       = null;
  for (x=1; x<=XLAST; x++)
  {
    Boxes[x][0].type = topedge;
    Boxes[x][0].avail = TRUE;
  }
  Boxes[XLAST+1][0].type = null;

  /* initialize inner rows */
  for (y=1; y<=YLAST; y++)
  {
    Boxes[0][y].type = leftedge;
    Boxes[0][y].avail = TRUE;
    for (x=1; x<=XLAST; x++)
      Boxes[x][YLAST+1].type = empty;
    Boxes[XLAST+1][y].type = rightedge;
    Boxes[XLAST+1][y].avail = TRUE;
  }

  /* initialize last row */
  Boxes[0][YLAST+1].type       = null;
  for (x=1; x<=XLAST; x++)
  {
    Boxes[x][YLAST+1].type = botedge;
    Boxes[x][YLAST+1].avail = TRUE;
  }
  Boxes[XLAST+1][YLAST+1].type = null;

  /* Place */
  i=0;
  while (i<nObjects) {
    x = rand()%XLAST+1;
    y = rand()%YLAST+1;
    if (Boxes[x][y].type == empty)
    {
      Boxes[x][y].type = full;
      i++;
    }
  }

  /* Ray Trace All Edges */
  for (y=0; y<=YLAST+1; y++)
    for (x=0; x<=XLAST+1; x++)
      if (IsEdge(x,y)) {
        Boxes[x][y].RayX = x;
        Boxes[x][y].RayY = y;
        Boxes[x][y].RayType = raytrace(&Boxes[x][y].RayX, &Boxes[x][y].RayY);
      }

  return;
}

void TGameWindow::Paint( HDC PaintDC, PAINTSTRUCT& )
{
  DrawBoard(PaintDC);

  DrawBMP( PaintDC,  15, 10, RemainBMP);
  DrawBMP( PaintDC, 310, 10, ScoreBMP);

  WriteScore( PaintDC );

  if ( IsGameOver )
    DrawGameOver( PaintDC );
}

void TGameWindow::ResetGame( TMessage& )
{
  if (!IsGameStarted) return;

  ModifyMenu( GetMenu( HWindow ), IDM_STOP, MF_BYCOMMAND,
	      IDM_STOP, "&Stop" );

  InvalidateRect( HWindow, FALSE, TRUE );
  if ( SetTimer( HWindow, 1, 100, NULL ) == 0 )
  {
    MessageBox( HWindow, "No Timers Left", "X-Ray Error", MB_OK );
    exit( 1 );
  }
  Counter = 0;
  Tenths  = 0;
  Score = START_SCORE;
  nRemain = nObjects;
  nMarks = nObjects;

  InitBoxes();

  IsGameOver = FALSE;
  IsGameStarted = FALSE;
  IsGameAborted = FALSE;
}

void TGameWindow::SetupWindow( void )
{
  Cursor[CURSOR_LEFT]  = LoadCursor( GetApplication()->hInstance, "RAYL1" );
  Cursor[CURSOR_RIGHT] = LoadCursor( GetApplication()->hInstance, "RAYR1" );
  Cursor[CURSOR_TOP]   = LoadCursor( GetApplication()->hInstance, "RAYT1" );
  Cursor[CURSOR_BOT]   = LoadCursor( GetApplication()->hInstance, "RAYB1" );
  Cursor[CURSOR_MARK]  = LoadCursor( GetApplication()->hInstance, "MARKM" );
  Cursor[CURSOR_ORG]   = (HCURSOR) GetClassWord(HWindow, GCW_HCURSOR);

  GameOver   = LoadBitmap( GetApplication()->hInstance, "GAMEOVER" );
  ScoreBMP   = LoadBitmap( GetApplication()->hInstance, "SCORE" );
  RemainBMP  = LoadBitmap( GetApplication()->hInstance, "REMAIN" );
  Reflect    = LoadBitmap( GetApplication()->hInstance, "REFLECT" );
  Hit        = LoadBitmap( GetApplication()->hInstance, "HIT" );
  Deflect    = LoadBitmap( GetApplication()->hInstance, "DEFLECT" );
  Deflect2   = LoadBitmap( GetApplication()->hInstance, "DEFLECT2" );
  MarkA      = LoadBitmap( GetApplication()->hInstance, "MarkA" );
  MarkB      = LoadBitmap( GetApplication()->hInstance, "MarkB" );
  BoxBMP     = LoadBitmap( GetApplication()->hInstance, "BoxBMP" );
  BoxgBMP    = LoadBitmap( GetApplication()->hInstance, "BoxgBMP" );
  BoxBall    = LoadBitmap( GetApplication()->hInstance, "BOXBALL");
  BoxBallX   = LoadBitmap( GetApplication()->hInstance, "BOXBALLX");
  MarkBX     = LoadBitmap( GetApplication()->hInstance, "MARKBX");

  GetParms();
  InitBoxes();

  IsGameOver = FALSE;
  IsGameStarted = FALSE;
  IsGameAborted = FALSE;
  Counter = 0;
  Tenths  = 0;
  Score = START_SCORE;
  nRemain = nObjects;
  nMarks = nObjects;
}

void TGameWindow::Stop( TMessage& )
{
  IsGameAborted = TRUE;
  StopGame( );
}

void TGameWindow::IDMSound( TMessage& )
{
  if (SoundOn)
  {
    SoundOn = FALSE;
    CheckMenuItem(GetMenu( HWindow ), IDM_SOUND, MF_UNCHECKED);
  }
  else
  {
    SoundOn = TRUE;
    CheckMenuItem(GetMenu( HWindow ), IDM_SOUND, MF_CHECKED);
   }
}

void TGameWindow::IDMTimer( TMessage& )
{
  if (TimerOn)
  {
    TimerOn = FALSE;
    CheckMenuItem(GetMenu( HWindow ), IDM_TIMER, MF_UNCHECKED);
  }
  else
  {
    TimerOn = TRUE;
    CheckMenuItem(GetMenu( HWindow ), IDM_TIMER, MF_CHECKED);
   }
}

void TGameWindow::IDMHelp( TMessage& )
{
  WinHelp( HWindow, ".\\XRAY.HLP", HELP_INDEX, 0L );
}

void TGameWindow::IDMHelpHelp( TMessage& )
{
  WinHelp( HWindow, "WINHELP.HLP", HELP_HELPONHELP, 0L );
}

void TGameWindow::GameExit( TMessage& Message )
{
  TGameWindow::WMDestroy(Message);
}

void TGameWindow::StopGame( )
{
  POINT Point;
  int i;
  char name[25];

  IsGameStarted = TRUE;

  /* Only stop the game once */
  if (IsGameOver) return;

  KillTimer( HWindow, 1 );
  WriteScore(0);

  ModifyMenu( GetMenu( HWindow ), IDM_STOP, MF_BYCOMMAND | MF_GRAYED,
	      IDM_STOP, "&Stop" );

  IsGameOver = TRUE;
  InvalidateRect( HWindow, FALSE, TRUE );

  /* return cursor to "normal" */
  SetClassWord( HWindow, GCW_HCURSOR, (WORD)Cursor[Current_cur=CURSOR_ORG]);
  GetCursorPos( &Point );
  SetCursorPos( Point.x, Point.y );

  /* Update High Scores */
  if (!IsGameAborted && Score > HighScores[4].score) {
    strcpy(name,"");
    if (GetApplication()->ExecDialog(new TInputDialog(this, "Best Score",
       "Enter your name:", name, sizeof(name))) == IDCANCEL)
    {
      strcpy(name,"Anonymous");
    }
    i = 4;
    while (i > 0 && Score > HighScores[i-1].score) {
      HighScores[i] = HighScores[i-1];
      i--;
    }
    HighScores[i].score = Score;
    strcpy(HighScores[i].name, name);
  }
}


void TGameWindow::WMDestroy( TMessage& Message )
{
  PutParms();

  DeleteObject( GameOver );
  DeleteObject( ScoreBMP );
  DeleteObject( RemainBMP );
  DeleteObject( Reflect );
  DeleteObject( Hit );
  DeleteObject( Deflect );
  DeleteObject( Deflect2);
  DeleteObject( MarkA );
  DeleteObject( MarkB );
  DeleteObject( BoxBMP );
  DeleteObject( BoxgBMP );
  DeleteObject( BoxBall );
  DeleteObject( BoxBallX);
  DeleteObject( MarkBX);

  WinHelp( HWindow, "XRAY.HLP", HELP_QUIT, 0L );
  StopMCI ();

  KillTimer( HWindow, 1 );
  TWindow::WMDestroy( Message );
}


void TGameWindow::WMMouseMove(TMessage& Message)
{
  POINT Box,Point;

  if ( IsGameOver)
    return;
  
  Box = FindBox(Message.LP.Lo, Message.LP.Hi);
  
  if (Boxes[Box.x][Box.y].type == topedge) {
    if (Current_cur != CURSOR_TOP) {
      Current_cur = CURSOR_TOP;
      SetClassWord( HWindow, GCW_HCURSOR, (WORD)Cursor[CURSOR_TOP]);
      GetCursorPos( &Point );
      SetCursorPos( Point.x, Point.y );
    }
  }
  else if (Boxes[Box.x][Box.y].type == botedge) {
    if (Current_cur != CURSOR_BOT) {
      Current_cur = CURSOR_BOT;
      SetClassWord( HWindow, GCW_HCURSOR, (WORD)Cursor[CURSOR_BOT]);
      GetCursorPos( &Point );
      SetCursorPos( Point.x, Point.y );
    }
  }
  else if (Boxes[Box.x][Box.y].type == rightedge) {
    if (Current_cur != CURSOR_RIGHT) {
      Current_cur = CURSOR_RIGHT;
      SetClassWord( HWindow, GCW_HCURSOR, (WORD)Cursor[CURSOR_RIGHT]);
      GetCursorPos( &Point );
      SetCursorPos( Point.x, Point.y );
    }
  }
  else if (Boxes[Box.x][Box.y].type == leftedge) {
    if (Current_cur != CURSOR_LEFT) {
      Current_cur = CURSOR_LEFT;
      SetClassWord( HWindow, GCW_HCURSOR, (WORD)Cursor[CURSOR_LEFT]);
      GetCursorPos( &Point );
      SetCursorPos( Point.x, Point.y );
    }
  }
  else if ((Boxes[Box.x][Box.y].type == empty) ||
           (Boxes[Box.x][Box.y].type == full)) {
    if (Current_cur != CURSOR_MARK) {
      Current_cur = CURSOR_MARK;
      SetClassWord( HWindow, GCW_HCURSOR, (WORD)Cursor[CURSOR_MARK]);
      GetCursorPos( &Point );
      SetCursorPos( Point.x, Point.y );
    }
  } 
  else  {
    if (Current_cur != CURSOR_ORG) {
      Current_cur = CURSOR_ORG;
      SetClassWord( HWindow, GCW_HCURSOR, (WORD)Cursor[CURSOR_ORG]);
      GetCursorPos( &Point );
      SetCursorPos( Point.x, Point.y );
    }
  } 
  return;
}

void TGameWindow::WMLButtonDown( TMessage& Message )
{
 POINT Box;

 if (!IsGameStarted)
   IsGameStarted = TRUE;

 if ( IsGameOver)
   return;

 Box = FindBox(Message.LP.Lo, Message.LP.Hi);
 if (Boxes[Box.x][Box.y].avail)
 {
   if (SoundOn) {
     if (Boxes[Box.x][Box.y].RayType == hit)
       PlayWave(1);
     else if (Boxes[Box.x][Box.y].RayType == deflect)
       PlayWave(2);
     else  /* reflect */
       PlayWave(3);
   }
   Boxes[Box.x][Box.y].avail = FALSE;
   if (Boxes[Box.x][Box.y].RayType == deflect)
   {
     Boxes[Boxes[Box.x][Box.y].RayX][Boxes[Box.x][Box.y].RayY].RayType = deflect2;
     Boxes[Boxes[Box.x][Box.y].RayX][Boxes[Box.x][Box.y].RayY].avail = FALSE;
     DrawEdge(0,Boxes[Box.x][Box.y].RayX,Boxes[Box.x][Box.y].RayY);
   }
   Score -= SCORE_RAY;
   if (Score < 0) Score = 0;
   WriteScore( 0 );
   if (Score == 0)
     StopGame();

   DrawEdge(0,Box.x,Box.y);

 }

 if (Boxes[Box.x][Box.y].mark > 0) {

   switch (Boxes[Box.x][Box.y].mark)
   {
     case 1:
	      if (nMarks > 0) {
		nMarks--;
		WriteScore(0);
		Boxes[Box.x][Box.y].mark = 2;
		DrawBox( 0, Boxes[Box.x][Box.y].UL.x,
	                    Boxes[Box.x][Box.y].UL.y, MarkB, 0 );
		if (Boxes[Box.x][Box.y].type == full) {
		   nRemain--;
		   if (nRemain == 0) {
		      if (SoundOn)
			PlayWave(4);
		      StopGame();
                   }
		}
	      }
                     else {
                /* only question marks left */
		Boxes[Box.x][Box.y].mark = 3;
		DrawBox( 0, Boxes[Box.x][Box.y].UL.x,
			     Boxes[Box.x][Box.y].UL.y, MarkA, 0 );
	      } 
	      break;
      case 2:
		Boxes[Box.x][Box.y].mark = 3;
		DrawBox( 0, Boxes[Box.x][Box.y].UL.x,
			     Boxes[Box.x][Box.y].UL.y, MarkA, 0 );
		if (Boxes[Box.x][Box.y].type == full) {
		  nRemain++;
		}
		nMarks++;
  	        WriteScore(0);
		break;
      case 3:
		/* Erase mark */
      default:
	       Boxes[Box.x][Box.y].mark=1;
	       DrawBox( 0, Boxes[Box.x][Box.y].UL.x,
	                    Boxes[Box.x][Box.y].UL.y, BoxBMP, 0 );
     }
  }
}

void TGameWindow::WMTimer( TMessage& )
{
  if (TimerOn && IsGameStarted)
    if (++Tenths > 9) {
      Counter++;
      if ((Score -= SCORE_SECOND) < 0) Score = 0;
      WriteScore( 0 );
      Tenths = 0;
    }
  if ( Score == 0 )
    StopGame();
}

void TGameWindow::WMSize( TMessage& )
{
  if ( IsGameOver )
    return;
  if ( IsIconic( HWindow ) )
    KillTimer( HWindow, 1 );
  else
    if ( SetTimer( HWindow, 1, 100, NULL ) == 0 )
    {
      MessageBox( HWindow, "No Timers Left","X-Ray Error", MB_OK );
      exit( 1 );
    }
}

void TGameWindow::WriteScore( HDC DC )
{
  char s[ 21 ];
  BOOL MadeDC;
  HBRUSH hbrush;
  HPEN hpen;

  if (DC == 0)
  {
    MadeDC = TRUE;
    DC = GetDC( HWindow );
  }
  else
    MadeDC = FALSE;

//  SelectObject( DC, CreateSolidBrush( 0x8080 ) );
  hbrush = (HBRUSH)GetStockObject( LTGRAY_BRUSH );
  hpen   = CreatePen(PS_SOLID, 1, RGB(192,192,192));

  SelectObject( DC, hbrush );
  SelectObject( DC, hpen );
  SetBkColor( DC, RGB(192,192,192));
//  SetBkMode( DC, OPAQUE );

  // Number of Objects left to Mark
  Rectangle( DC, 20, 30, 52, 50 );
  sprintf( s, "%4d", nMarks );
  TextOut( DC, 20, 30, s, strlen( s ) );

  // Score
  Rectangle( DC, 310, 30, 350, 50 );
  sprintf( s, "%4d", Score );
  TextOut( DC, 310, 30, s, strlen( s ) );


//  DeleteObject( SelectObject( DC, GetStockObject( GRAY_BRUSH ) ) );
//  SelectObject( DC, GetStockObject( NULL_PEN ) );

  DeleteObject(hbrush);
  DeleteObject(hpen);

  if ( MadeDC )
    ReleaseDC( HWindow, DC );
}

// Play a *.wav file
void TGameWindow::PlayWave ( int wave )
{
  if (wave == 1)
    strcpy(elementName,".\\HIT.WAV");
  else if (wave == 2)
    strcpy(elementName,".\\DEFLECT.WAV"); 
  else if (wave == 3)
    strcpy(elementName,".\\REFLECT.WAV"); 
  else if (wave == 4)
    strcpy(elementName,".\\GAMEOVER.WAV"); 

  if (!bRunning) {
    // MCI APIs to open a device and play a .WAV file, using
    // notification to close
    MciOpenParm.dwCallback = 0L;
    MciOpenParm.wDeviceID = 0;
    MciOpenParm.wReserved0 = 0;
    MciOpenParm.lpstrDeviceType = NULL;
    MciOpenParm.lpstrElementName = elementName;
    MciOpenParm.lpstrAlias = NULL;

    if (mciSendCommand (0, MCI_OPEN, (DWORD) (MCI_WAIT | MCI_OPEN_ELEMENT),
                             (DWORD)(LPMCI_OPEN_PARMS)&MciOpenParm)) {
      MessageBox (HWindow, "Error Opening Sound (WAV) file.\n Sounds disabled.",
	"X-Ray Wave Player Error", MB_OK);
      SoundOn = FALSE;
      CheckMenuItem(GetMenu( HWindow ), IDM_SOUND, MF_UNCHECKED);
      return;
    } 

    wDeviceID = MciOpenParm.wDeviceID;

    // The time format in this demo is in Samples.
//    MciSetParm.dwCallback = 0L;
//    MciSetParm.dwTimeFormat = MCI_FORMAT_SAMPLES;
//    if (mciSendCommand (wDeviceID, MCI_SET, MCI_SET_TIME_FORMAT,
//                         (DWORD)(LPMCI_SET_PARMS)&MciSetParm)) {
//         MessageBox (HWindow, "SetTime Error", "Sound Play", MB_OK);
//	 return;
//      }

      MciPlayParm.dwCallback = (unsigned long)HWindow;
      MciPlayParm.dwFrom = 0;
      MciPlayParm.dwTo = 0;
      mciSendCommand (wDeviceID, MCI_PLAY, MCI_NOTIFY,
                      (DWORD) (LPMCI_PLAY_PARMS)&MciPlayParm);
      bRunning = TRUE;

   }
}

void TGameWindow::StopMCI ()
{
  MciGenParm.dwCallback = 0L;
  mciSendCommand (wDeviceID, MCI_STOP, MCI_WAIT,
                  (DWORD)(LPMCI_GENERIC_PARMS)&MciGenParm);
  mciSendCommand (wDeviceID, MCI_CLOSE, MCI_WAIT,
                  (DWORD)(LPMCI_GENERIC_PARMS)&MciGenParm);

  bRunning = FALSE;
  wDeviceID = 0;
}

// Response function MM_MCINOTIFY message when MCI_PLAY is complete.
void TGameWindow::MciNotify (RTMessage)
{
   if (wDeviceID) 
     StopMCI ();
}



// --------------- TApp ------------------------

void TApp::InitMainWindow( void )
{
  MainWindow = new TGameWindow( NULL, "X-Ray" );
}

void TApp::InitInstance( void )
{
  TApplication::InitInstance();
  HAccTable = LoadAccelerators( hInstance, MAKEINTRESOURCE( XRAYACCEL ) );
}

// -------------Main Program--------------------

int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
		   LPSTR lpCmd, int nCmdShow)
{
  TApp App( "X-RayGame", hInstance, hPrevInstance, lpCmd, nCmdShow);
  App.Run();
  return ( App.Status );
}



